﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace UnattendedExample
{
    internal class Program
    {
        /// <summary>
        /// Загрузка банковских заявлений и извлечение ФИО клиента
        /// </summary>
        /// <param name="args"></param>
        private static void Main(string[] args)
        {
            var filename = "Banking_eng.pdf";
            if (args != null && args.Length > 0)
            {
                filename = args[0];
            }
            Console.WriteLine("recognition of " + filename);
            PrivateDocumentsRecognition(filename);
            Console.WriteLine("press any key to exit...");
            Console.ReadKey();
        }

        /// <summary>
        /// Обработка нескольких изображений
        /// <param name="filename">Имя обрабатываемого файла</param>
        /// </summary>
        private static void PrivateDocumentsRecognition(string filename)
        {
            // FlexiCaptureWebServiceSoapClient - имя интерфейса, реализованного web-сервисом
            // FlexiCapture.
            // Создаем URL для ссылки на третью версию службы.
            var service = new FlexiCapture.FlexiCaptureWebServiceSoapClient();

            // ======= BASIC AUTHENTICATION =======
            // service.ClientCredentials.UserName.UserName = "username";
            // service.ClientCredentials.UserName.Password = "password";

            // Для обеспечения приватности создаваемых пакетов, нужно указать текущего пользователя
            // Сначала получаем имя пользователя, видимое системой
            var username = service.GetCurrentUserIdentity();
            // Затем ищем себя среди пользователей FlexiCapture
            var userId = service.FindUser(username.Name);
            if (userId <= 0) throw new Exception("Текущий пользователь не найден");

            Console.WriteLine("connected to FlexiCapture, user id: " + userId);

            // Открываем сессию и начинаем обработку
            OpenWorkSession(service, userId, filename);
        }

        /// <summary>
        /// Открытие сессии и начало обработки
        /// </summary>
        /// <param name="service">Сервис FlexiCapture</param>
        /// <param name="userId">Идентификатор пользователя</param>
        /// <param name="filename">Имя обрабатываемого файла</param>
        private static void OpenWorkSession(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int userId, string filename)
        {
            // Параметры сессии
            const int roleType = 11; //Роль оператора пользовательской станции
            const int stationType = 10; //Пользовательская станция

            // Открываем новую сессию обработки
            var sessionId = service.OpenSession(roleType, stationType);
            if (sessionId <= 0) throw new Exception("Не удалось открыть сессию");
            try
            {
                // Открываем первый пакет и продолжаем обработку
                OpenBankingProject(service, userId, sessionId, filename);
            }
            finally
            {
                // Открытую сессию обязательно нужно закрыть по завершении работы
                service.CloseSession(sessionId);
            }
        }

        /// <summary>
        /// Открытие первого найденного проекта и продолжение обработки
        /// </summary>
        /// <param name="service">Сервис FlexiCapture</param>
        /// <param name="userId">Идентификатор пользователя</param>
        /// <param name="sessionId">Идентификатор сессии</param>
        /// <param name="filename">Имя обрабатываемого файла</param>
        private static void OpenBankingProject(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int userId, int sessionId, string filename)
        {
            // Получаем список проектов (необязательно при работе с
            // одним проектом, в этом случае удобнее
            // открывать проект с использованием GUID)
            var projects = service.GetProjects();
            var projectGuid = "";
            if (projects != null && projects.Count > 0)
            {
                foreach (var project in projects)
                {
                    if (project.Name != "UnattendedExample") continue;
                    projectGuid = project.Guid;
                    break;
                }
            }

            if (string.IsNullOrEmpty(projectGuid))
            {
                throw new Exception("Проект UnattendedExample не найден. Загрузите его на сервер перед использованием примера");
            }

            // Открываем первый проект с именем UnattendedExample в списке
            var projectId = service.OpenProject(sessionId, projectGuid);
            if (projectId <= 0) throw new Exception("Не удалось открыть проект");
            try
            {
                // Создаем новый пакет и продолжаем обработку
                ProcessRecognition(service, userId, sessionId, projectId, filename);
            }
            finally
            {
                // Открытый пакет нужно закрыть перед завершением работы
                service.CloseProject(sessionId, projectId);
            }
        }

        /// <summary>
        /// Обработка изображений
        /// </summary>
        /// <param name="service">Сервис FlexiCapture</param>
        /// <param name="userId">Идентификатор пользователя</param>
        /// <param name="sessionId">Идентификатор сессии</param>
        /// <param name="projectId">Идентификатор проекта</param>
        /// <param name="filename">Имя обрабатываемого файла</param>
        private static void ProcessRecognition(FlexiCapture.
        FlexiCaptureWebServiceSoapClient service, int userId, int sessionId, int projectId, string filename)
        {
            // Создаем новый пакет. Необходимо указать имя нового пакета, все остальные
            // свойства остаются по-умолчанию
            var batch = new FlexiCapture.Batch { Name = "Sample API Batch" };
            var batchId = service.AddNewBatch(sessionId, projectId, batch, userId);
            if (batchId <= 0) throw new Exception("Не удалось создать пакет");
            try
            {
                AddImageToBatch(service, sessionId, projectId, batchId, filename);

                // Запускаем обработку пакета
                service.ProcessBatch(sessionId, batchId);
                Console.WriteLine("recognition task created, waiting ");

                // Ожидаем завершение обработки
                var percentCompleted = 0;
                while (percentCompleted < 100)
                {
                    Console.CursorLeft = 0;
                    Console.Write(percentCompleted + "%");
                    percentCompleted = service.GetBatchPercentCompleted(batchId);
                    System.Threading.Thread.Sleep(500);
                }
                Console.CursorLeft = 0;
                Console.WriteLine("complete...");

                // Получаем результаты
                var documents = service.GetDocuments(sessionId, batchId);
                if (documents == null) return;

                // XML имена, требуются для разбора XML-результатов
                var docs = XName.Get("Documents", "http://www.abbyy.com/FlexiCapture/Schemas/Export/FormData.xsd");
                var banking = XName.Get("_Banking_eng", "http://www.abbyy.com/FlexiCapture/Schemas/Export/Banking_eng.xsd");

                // Перебираем все результаты и выводим их на экран
                // Из каждого документа выводим только три поля:
                // обращение (Addressing), имя (First Name), фамилия (Last Name)
                foreach (var document in documents)
                {
                    if (document.Id == 0) continue;
                    // Получаем XML-файл с результатами распознавания
                    var attachedFile = service.LoadDocumentResult(sessionId, batchId, document.Id, "Result.xml");
                    if (attachedFile.Bytes == null) continue;
                    // Открываем результат в формате XML
                    var xml = XDocument.Load(new MemoryStream(attachedFile.Bytes));
                    // Контейнер распознанных данных
                    var docsElement = xml.Element(docs);
                    if (docsElement == null) continue;
                    var result = docsElement.Element(banking);
                    if (result == null) continue;
                    // Получаем распознанные данные ФИО
                    var addressing = result.Element("_Addressing");
                    var surname = result.Element("_Last_Name");
                    var name = result.Element("_First_Name");
                    // Выводим данные на экран
                    Console.WriteLine("- " +
                        (addressing == null ? "" : addressing.Value) + " " + 
                        (name == null ? "" : name.Value) + " " + 
                        (surname == null ? "" : surname.Value)
                        );
                }
            }
            finally
            {
                // Созданный пакет после обработки не нужен, удаляем
                service.DeleteBatch(sessionId, batchId);
            }
        }

        /// <summary>
        /// Заполнение созданного пакета изображениями
        /// </summary>
        /// <param name="service">Сервис FlexiCapture</param>
        /// <param name="sessionId">Идентификатор сессии</param>
        /// <param name="projectId">Идентификатор проекта</param>
        /// <param name="batchId">Идентификатор пакета</param>
        /// <param name="filename">Имя обрабатываемого файла</param>
        private static void AddImageToBatch(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int sessionId, int projectId, int batchId, string filename)
        {
            // Открываем пакет перед добавленем изображений
            service.OpenBatch(sessionId, batchId);
            try
            {
                // Простая пересылка для файлов < 256кб
                if (new FileInfo(filename).Length < 0x40000)
                {
                    // Для маленьких файлов есть довольно простой способ загрузить их на файл
                    // Тот же способ можно использовать и для больших файлов при условии хорошей сети.
                    // При этом стоит помнить, что в таком случае данные передются в кодировке
                    // base64, что увеличивает нагрузку на сеть на 33%
                    service.AddNewImage(sessionId, batchId, new FlexiCapture.File()
                    {
                        Bytes = File.ReadAllBytes(filename),
                        Name = filename
                    });
                }
                else
                {
                    // Если файл превыщает установленный размер, то его пересылка целиком в base64 имеет ряд недостатков:
                    // увеличение нагрузки на сеть на 33%, возможность блокировки слишком тяжёлого запроса 
                    // IIS-ом или межсетевыми экранами, необходимость пересылать весь файл заново в случае разрыва
                    // соединения или ошибок в сети.
                    // Способ пересылки через API файлового серсиса оказывается более эффективным методом.
                    var doc = new FlexiCapture.Document { BatchId = batchId };
                    var file = new FlexiCapture.File { Name = filename };
                    var documentId = service.AddNewDocument(sessionId, doc, file, false, 0);
                    UploadFile(service, sessionId, projectId, batchId, documentId, filename).Wait();
                }
            }
            finally
            {
                service.CloseBatch(sessionId, batchId);
            }
        }

        /// <summary>
        /// Передача файла
        /// </summary>
        /// <param name="service">Сервис FlexiCapture</param>
        /// <param name="sessionId">Идентификатор сессии</param>
        /// <param name="projectId">Идентификатор проекта</param>
        /// <param name="batchId">Идентификатор пакета</param>
        /// <param name="documentId">Идентификатор документа</param>
        /// <param name="filename">Имя обрабатываемого файла</param>
        /// <returns></returns>
        private static async Task UploadFile(FlexiCapture.FlexiCaptureWebServiceSoapClient service, int sessionId, int projectId,
            int batchId, int documentId, string filename)
        {
            const int objectType = 0;
            var crc = new Crc32();
            using (var fs = File.OpenRead(filename))
            {
                // Передаём частями по 1 МБ
                var buffer = new byte[0x100000];

                var readed = fs.Read(buffer, 0, buffer.Length);
                var checksum = crc.Next(buffer, 0, readed);
                if (readed < buffer.Length)
                {
                    // Маленькие файлы грузятся одним куском
                    var resp = await FileRequest(service, "Save", objectType, sessionId, projectId, batchId, 0, documentId, 0, filename, 0,
                        new ByteArrayContent(buffer, 0, readed));
                    if (resp.StatusCode != HttpStatusCode.OK) throw new Exception("Ошибка сервера");
                }
                else
                {
                    var offset = 0;
                    while (readed > 0)
                    {
                        var action = offset == 0 ? "BeginSaveChunked" : "Append";
                        await FileRequest(service, action, objectType, sessionId, projectId, batchId, 0, documentId, 0, filename,
                            offset, new ByteArrayContent(buffer, 0, readed));
                        offset += readed;
                        readed = fs.Read(buffer, 0, buffer.Length);
                        checksum = crc.Next(buffer, 0, readed);
                    }
                    var resp = await FileRequest(service, "Commit", objectType, sessionId, projectId, batchId, 0, documentId, 0, filename);
                    if (resp.StatusCode != HttpStatusCode.OK) throw new Exception("Ошибка сервера");
                }
                var response = await FileRequest(service, "Checksum", objectType, sessionId, projectId, batchId, 0, documentId, 0, filename);
                var text = await response.Content.ReadAsStringAsync();
                if (uint.Parse(text, NumberStyles.HexNumber) != checksum)
                {
                    throw new Exception("Файл закачан некорректно");
                }
            }
        }

        /// <summary>
        /// Отправка запроса файловому сервису FlexiCapture
        /// </summary>
        /// <remarks>Данные кодируются как x-www-form-urlencoded, что экономит траффик</remarks>
        /// <param name="service">Сервис</param>
        /// <param name="action">Запрашиваемое дествие</param>
        /// <param name="objectType">Тип обрабатываемого объекта</param>
        /// <param name="sessionId">Идентификатор сессии</param>
        /// <param name="projectId">Идентификатор проекта</param>
        /// <param name="batchId">Идентификатор пакета</param>
        /// <param name="parentId">Идентификатор родителя объекта</param>
        /// <param name="objectId">Идентификатор обрабатываемого объекта</param>
        /// <param name="version">Версия обрабатываемого объекта</param>
        /// <param name="streamName">Имя потока</param>
        /// <returns>Http ответ от сервера</returns>
        private static async Task<HttpResponseMessage> FileRequest(FlexiCapture.FlexiCaptureWebServiceSoapClient service, string action,
            int objectType, int sessionId, int projectId, int batchId, int parentId, int objectId, int version, string streamName)
        {
            var creds = CredentialCache.DefaultNetworkCredentials;
            if (!string.IsNullOrEmpty(service.ClientCredentials.UserName.UserName))
                creds = new NetworkCredential(service.ClientCredentials.UserName.UserName, service.ClientCredentials.UserName.Password);
            using (var handler = new HttpClientHandler { Credentials = creds })
            using (var client = new HttpClient(handler))
            {
                var uri = service.Endpoint.Address.Uri;
                var content = new FormUrlEncodedContent(new Dictionary<string, string>
                {
                    {"Action", action},
                    {"objectType", objectType.ToString()},
                    {"sessionId", sessionId.ToString()},
                    {"projectId", projectId.ToString()},
                    {"batchId", batchId.ToString()},
                    {"parentId", parentId.ToString()},
                    {"objectId", objectId.ToString()},
                    {"version", version.ToString()},
                    {"streamName", streamName},
                });
                return await client.PostAsync(uri.OriginalString.Replace("/API/v1/Soap", "/FileService/v1"), content);
            }
        }

        /// <summary>
        /// Отправка запроса файловому сервису FlexiCapture
        /// </summary>
        /// <remarks>Данные кодируются как multipart/form-data, что позволяет отправлять на сервер файлы</remarks>
        /// <param name="service">Сервис</param>
        /// <param name="action">Запрашиваемое дествие</param>
        /// <param name="objectType">Тип обрабатываемого объекта</param>
        /// <param name="sessionId">Идентификатор сессии</param>
        /// <param name="projectId">Идентификатор проекта</param>
        /// <param name="batchId">Идентификатор пакета</param>
        /// <param name="parentId">Идентификатор родителя объекта</param>
        /// <param name="objectId">Идентификатор обрабатываемого объекта</param>
        /// <param name="version">Версия обрабатываемого объекта</param>
        /// <param name="streamName">Имя потока</param>
        /// <param name="offset">Смещение файла</param>
        /// <param name="file">Передаваемый файл</param>
        /// <returns>Http ответ от сервера</returns>
        private static async Task<HttpResponseMessage> FileRequest(FlexiCapture.FlexiCaptureWebServiceSoapClient service, string action,
            int objectType, int sessionId, int projectId, int batchId, int parentId, int objectId, int version, string streamName, int offset, HttpContent file)
        {
            var creds = CredentialCache.DefaultNetworkCredentials;
            if (!string.IsNullOrEmpty(service.ClientCredentials.UserName.UserName))
                creds = new NetworkCredential(service.ClientCredentials.UserName.UserName, service.ClientCredentials.UserName.Password);
            using (var handler = new HttpClientHandler { Credentials = creds })
            using (var client = new HttpClient(handler))
            {
                var uri = service.Endpoint.Address.Uri;
                var content = new MultipartFormDataContent
                {
                    {new StringContent(action), "Action"},
                    {new StringContent(objectType.ToString()), "objectType"},
                    {new StringContent(sessionId.ToString()), "sessionId"},
                    {new StringContent(projectId.ToString()), "projectId"},
                    {new StringContent(batchId.ToString()), "batchId"},
                    {new StringContent(parentId.ToString()), "parentId"},
                    {new StringContent(objectId.ToString()), "objectId"},
                    {new StringContent(version.ToString()), "version"},
                    {new StringContent(Convert.ToBase64String(Encoding.Unicode.GetBytes(streamName))), "streamName"},
                };
                if (offset > 0)
                {
                    content.Add(new StringContent(offset.ToString()), "offset");
                }
                content.Add(file, "blob", "data.txt");
                return await client.PostAsync(uri.OriginalString.Replace("/API/v1/Soap", "/FileService/v1"), content);
            }
        }

    }
}
